import logging
from collections import defaultdict
from typing import DefaultDict, Dict, List

from intervaltree import Interval, IntervalTree
from lsprotocol import types as lsp_types

logger = logging.getLogger(__name__)

References = DefaultDict[str, List[lsp_types.Location]]
Definitions = Dict[str, lsp_types.Location]
LocationToDefinition = Dict[lsp_types.Location, str]


class LocationKeyMap:
    def __init__(self):
        self.file_to_tree = defaultdict(IntervalTree)

    def add_location_key(self, position: lsp_types.Location, value: str) -> None:
        interval = Interval(
            (position.range.start.line, position.range.start.character),
            (position.range.end.line, position.range.end.character),
            value,
        )
        self.file_to_tree[position.uri].add(interval)

    def find_key_by_position(
        self, position: lsp_types.Position, doc_id: str
    ) -> str | None:
        logger.debug(
            f"Finding key for {position.line}-{position.character} in {doc_id}"
        )
        overlapping_intervals = self.file_to_tree[doc_id].at(
            (position.line, position.character)
        )

        for interval in overlapping_intervals:
            return interval.data

        return None

    def find_key_by_location(self, position: lsp_types.Location) -> str | None:
        return self.find_key_by_position(position.range.start, position.uri)


class HydraContext:
    """
    Stores: YAML keys and values pairs
    """

    def __init__(
        self,
        config: Dict,
        references: References = defaultdict(list),
        definitions: Definitions = {},
    ):
        self.config = config
        self.references = references
        self.definitions = definitions
        self.loc_to_definition = LocationKeyMap()
        for k, v in definitions.items():
            self.loc_to_definition.add_location_key(v, k)

    def set(self, key: str, value: str):
        raise NotImplementedError

    def get(self, key: str):
        """
        Get a value from the config:
            1. Using a standard dict notation (e.g. config["data"])
            2. Using a dot notation (e.g. config["data.loader"])

        Will return None if the key is not found
        """
        if "." not in key:
            return self.config.get(key)

        # nested case
        value = self.config
        for k in key.split("."):
            value = value.get(k)

            if value is None:
                return None

        return value



import json
import logging

import pygtrie
from lsprotocol.types import (
    CompletionItem,
    CompletionItemKind,
    CompletionList,
    CompletionParams,
    MarkupContent,
)
from pygls.server import LanguageServer

from hydra_lsp.context import HydraContext
from hydra_lsp.utils import to_markdown_content, yaml_get_var_prefix

logger = logging.getLogger(__name__)


class Completer:
    def __init__(self):
        self.context = None
        self.trie = pygtrie.CharTrie()

    def update(self, context: HydraContext):
        self.context = context
        self.trie.update(context.definitions)

    def get_completions(self, ls: LanguageServer, params: CompletionParams):
        uri = params.text_document.uri
        logger.info(f"Completion requested: {uri} at {params.position}")

        document = ls.workspace.get_document(uri)
        position = params.position
        current_line = document.lines[position.line]

        prefix = yaml_get_var_prefix(current_line, position.character)
        return self.complete(prefix)

    def complete(self, prefix: str | None) -> CompletionList:
        if prefix is None:
            return CompletionList(is_incomplete=False, items=[])

        try:
            keys = self.trie.keys(prefix)
        except KeyError:
            keys = []

        items = [
            CompletionItem(
                label=k,
                kind=CompletionItemKind.Variable,
                documentation=self._get_docstring(k),
            )
            for k in keys
        ]

        return CompletionList(is_incomplete=False, items=items)

    def _get_docstring(self, key: str) -> MarkupContent:
        value = str(self.context.get(key))

        s = json.dumps({key: value}, indent=2)[1:-1]
        result = to_markdown_content(s)

        return result



import collections
import logging

from lsprotocol.types import MarkupContent, MarkupKind

logger = logging.getLogger(__name__)


def to_markdown_content(value: str, lang: str = "json") -> MarkupContent:
    """Return the MarkupContent with Markdown kind."""
    return MarkupContent(kind=MarkupKind.Markdown, value=f"```{lang}\n{value}\n```")


def yaml_get_var_prefix(line: str, pos: int = 0) -> str | None:
    """
    Assume that the line should be in the following format:
        var: something ${va<cursor>xx} bbb
    it will return "va"
    """
    start = line.rfind("${", 0, pos)
    if start == -1:
        return None

    return line[start + len("${") : pos]


def yaml_get_key(line: str, position: int) -> str | None:
    """
    Get key from the yaml value (if exists).
    convert the line like this:
        "foo: foo ${bar}/something else" -> "bar"
    given the position of the cursor (in this case, it's on the "b", "a" or "r" letter)
    """
    start = line.find(":")
    if start == -1:
        return None

    if position > start:
        return None

    return line[:start].strip()


def yaml_get_variable_name(line: str, position: int) -> str | None:
    """
    Get variable name from the yaml value (if exists).
    convert the line like this:
        "foo: foo ${bar}/something else" -> "bar"
    given the position of the cursor (in this case, it's on the "b", "a" or "r" letter)
    """
    start = line.find(":")
    if start == -1:
        return None

    first_part_index = line.rfind("${", start, position)
    if first_part_index == -1:
        return None

    second_part_index = line.find("}", position + 1)
    if second_part_index == -1:
        return None

    return line[first_part_index + 2 : second_part_index]


def deep_update(source, overrides):
    """
    Update a nested dictionary or similar mapping.
    Modify ``source`` in place.
    """
    for key, value in overrides.items():
        if isinstance(value, collections.abc.Mapping) and value:
            source[key] = deep_update(source.get(key, {}), value)
        else:
            source[key] = value
    return source



import json
import logging

from lsprotocol import types as lsp_types
from pygls.server import LanguageServer

from hydra_lsp.parser import HydraContext
from hydra_lsp.utils import to_markdown_content, yaml_get_key, yaml_get_variable_name

logger = logging.getLogger(__name__)


# TODO: this class has a lot of duplicated code
class HydraIntel:
    def __init__(self, ls: LanguageServer) -> None:
        self.ls = ls

    def get_hover(
        self,
        params: lsp_types.HoverParams,
        context: HydraContext | None,
        debug_hover: bool = False,
    ) -> lsp_types.Hover | None:
        """
        Get hover information:

        - if the cursor is on the variable (after ':' and between '${' and '}'),
            it will return the value of the variable.
        - if the cursor is on the key (before ':'),
            it will return the computed value of the key.
        """

        document_path = params.text_document.uri
        document = self.ls.workspace.get_document(document_path)
        position = params.position

        if debug_hover:
            logger.info(f"Hover requested: {document_path} at {position}")

        if context is None:
            logger.warning("Context is not loaded")
            return None

        current_line = document.lines[position.line]

        # check if the cursor is before or after the ':'
        if position.character > current_line.find(":"):
            key = yaml_get_variable_name(current_line, position.character)
        else:
            key = context.loc_to_definition.find_key_by_position(
                position, document_path
            )
            logger.info(f"Key from position: {key}")

        value = context.get(key) if key is not None else None

        if key is None or value is None:
            return None

        s = json.dumps({key: value}, indent=2)[1:-1]
        return lsp_types.Hover(contents=to_markdown_content(s))

    # TODO: get_definition and get_references are very similar, we should merge them somehow
    def get_definition(
        self,
        params: lsp_types.TextDocumentPositionParams,
        context: HydraContext | None,
    ):
        """Get definition of the variable."""

        logger.info(
            f"Definition requested: {params.text_document.uri} at {params.position}"
        )
        document_path = params.text_document.uri
        document = self.ls.workspace.get_document(document_path)
        position = params.position

        current_line = document.lines[position.line]
        key = yaml_get_variable_name(current_line, position.character)

        if context is None:
            logger.warning("Context is not loaded")
            return None

        if key is None:
            return None

        logger.info(f"Definition of {key} is {context.definitions.get(key)}")
        return context.definitions.get(key)

    def get_references(
        self, params: lsp_types.ReferenceParams, context: HydraContext | None
    ):
        """Get references of the variable."""

        logger.info(
            f"References requested: {params.text_document.uri} at {params.position}"
        )
        document_path = params.text_document.uri
        document = self.ls.workspace.get_document(document_path)
        position = params.position

        current_line = document.lines[position.line]
        key = yaml_get_variable_name(current_line, position.character)

        if context is None:
            logger.warning("Context is not loaded")
            return None

        if key is None:
            key = yaml_get_key(current_line, position.character)

        if key is None:
            return None

        logger.info(f"References of {key} are {context.references.get(key)}")
        return context.references.get(key)

    def get_diagnostics(self, context: HydraContext | None, doc_uri: str | None):
        """Get diagnostics for the current context."""

        if context is None:
            logger.warning("Context is not loaded")
            return None

        diagnostics = []

        for reference, locations in context.references.items():
            if reference in context.definitions:
                continue

            for loc in locations:
                if doc_uri is not None and loc.uri != doc_uri:
                    continue

                diagnostics.append(
                    lsp_types.Diagnostic(
                        range=loc.range,
                        message=f"`{reference}` is not defined",
                        source="hydra-lsp",
                    )
                )

        logger.debug(f"Diagnostics: {diagnostics}")
        return diagnostics
